--============================================================================================
--##################################   Module Information   ##################################
--============================================================================================
--
-- Company: CERN 
--
-- Language: VHDL 2008
--
-- Description:
--
--        See adc_4ch_ad7682 Module
--
------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


------------------------------------------------------------------------------------------------------
-- Entity
------------------------------------------------------------------------------------------------------
entity adc_4ch_ad7682_spi_protocol is
  generic(
    --synthesis or simuation mode
    g_SYNTHESIS               : integer  := 1;   -- 1=true / 0=false
    -- Refresh period    
    g_WAIT_GAP_BEFORE_REFRESH : integer  := 0; -- clock cycles between the next conversion         
    -- Local slave bus
    g_CMB_SPI_ADDR_LEN        : positive := 1; -- slave selection (hot bit)
    g_CMB_SPI_DATA_LEN        : positive := 16 -- SPI frame length in bits
  );
  port(
    --clock and reset
    rst_ir                    : in  std_logic; 
    clk_ik                    : in  std_logic; 
    cen_ie                    : in  std_logic; 
    --enable
    enable_i                  : in  std_logic; 
    --timer
    timer_1us_i               : in  std_logic;
    --ADC CTRL
    ADC3_LV_volt_valid_o      : out std_logic;
    ADC3_LV_volt_ob16         : out std_logic_vector(15 downto 0);
    ADC3_HV1_curr_valid_o     : out std_logic;
    ADC3_HV1_curr_ob16        : out std_logic_vector(15 downto 0); 
    ADC3_HV2_curr_valid_o     : out std_logic;
    ADC3_HV2_curr_ob16        : out std_logic_vector(15 downto 0); 
    ADC3_LV_curr_valid_o      : out std_logic;
    ADC3_LV_curr_ob16         : out std_logic_vector(15 downto 0); 
    ADC3_4ch_temp_valid_o     : out std_logic;
    ADC3_4ch_temp_ob16        : out std_logic_vector(15 downto 0); 
    -- CMB SPI output
    cmb_frame_o               : out std_logic;
    cmb_addr_ob               : out std_logic_vector(g_CMB_SPI_ADDR_LEN-1 downto 0); --slave selection
    cmb_enable_ob             : out std_logic_vector(g_CMB_SPI_DATA_LEN-1 downto 0);
    cmb_wr_o                  : out std_logic;
    cmb_wr_data_ob            : out std_logic_vector(g_CMB_SPI_DATA_LEN-1 downto 0); --shift reg content
    cmb_rd_o                  : out std_logic;
    cmb_rd_data_ib            : in  std_logic_vector(g_CMB_SPI_DATA_LEN-1 downto 0);
    cmb_rd_valid_i            : in  std_logic;
    cmb_busy_i                : in  std_logic
  );
end entity ;


------------------------------------------------------------------------------------------------------
-- Architecture
------------------------------------------------------------------------------------------------------
architecture rtl of adc_4ch_ad7682_spi_protocol is
  
  --log2 function
  function log2(x : positive) return natural is
     constant log2_max : natural  := natural'high;
     variable log2_x   : natural  := 0;
     variable x_approx : positive := 1;
  begin
     for log2_x in 0 to log2_max loop
        if ( x_approx < x ) then
           x_approx := 2*x_approx;
        else
           return log2_x;
        end if;
     end loop;
     return log2_max;
  end log2;
  
  --get conversion time function
  function get_start_time return unsigned is
    variable v_start_time : integer := 200000;
  begin
    if g_SYNTHESIS>0 then
      v_start_time := 200000;  -- 200ms 
    else
      v_start_time := 10;      -- 10us to speed up simulation
    end if; 
    return to_unsigned(v_start_time, 32); -- conversion to unsigned
  end function;
  
  --get conversion time function
  function get_conv_time return unsigned is
    variable v_conv_time : integer := 100;
  begin
    if g_SYNTHESIS>0 then
      v_conv_time := 100;    -- 100us (min CONV time 5us)
    else
      v_conv_time := 1;      -- 1us to speed up simulation
    end if; 
    return to_unsigned(v_conv_time, 32); -- conversion to unsigned
  end function;

  --Conversion duration
  constant c_START_DUR_US     : unsigned(31 downto 0) := get_start_time; -- time at startup
  constant c_CONV_DUR_US      : unsigned(31 downto 0) := get_conv_time;  -- conversion time (min CONV time 5us)
  signal conv_wait_cnt_u      : unsigned(31 downto 0);

  --refresh gap counter
  signal refresh_gap_cnt_u    : unsigned(log2(g_WAIT_GAP_BEFORE_REFRESH+1)+1 downto 0);

  --conv channel counter
  constant c_CH_NUMBER        : integer := 4;         --4 channels 
  signal conv_ch_cnt          : unsigned(3 downto 0); --4 channels + temperature

  -- CMB internal
  signal cmb_wr_rd            : std_logic;
  signal cmb_wr_data_b        : std_logic_vector(g_CMB_SPI_DATA_LEN-1 downto 0);

  -- SPI cycle FSM :
  type t_state is
  (s_STARTUP,
   s_IDLE,
   s_CONV,
   s_CONV_WAIT);
  attribute syn_enum_encoding : STRING;
  attribute syn_enum_encoding of t_state : type is "one-hot";
  signal state                : t_state;
  
  --new cycle of conversion
  signal start_spi_cycle      : std_logic;

begin 
  
  -----------------
  -- SPI gap between refresh
  -----------------
  p_refresh_gap: process(clk_ik, rst_ir)
  begin
    if rst_ir = '1' then
     refresh_gap_cnt_u <= (others=>'0'); --do not wait at startup
     start_spi_cycle   <= '1';

    elsif rising_edge(clk_ik) then
      if cen_ie = '1' then
        if state = s_IDLE then
          if refresh_gap_cnt_u = 0 then 
            start_spi_cycle   <= '1';
          else
            start_spi_cycle   <= '0';
            refresh_gap_cnt_u <= refresh_gap_cnt_u - 1;
          end if;
        else
          start_spi_cycle     <= '0';
          if g_SYNTHESIS>0 then
            refresh_gap_cnt_u <= to_unsigned(g_WAIT_GAP_BEFORE_REFRESH, refresh_gap_cnt_u'length);
          else
            refresh_gap_cnt_u <= to_unsigned(10,refresh_gap_cnt_u'length); --to speed up simulation
          end if;
        end if;
      end if;
    end if;
  end process;


  ----------------------------------------------------------------------------------------------------
  -- SPI state machine
  ----------------------------------------------------------------------------------------------------
  p_spi_FSM : process(rst_ir, clk_ik)
  begin
    if rst_ir = '1' then
      cmb_wr_rd       <= '0';
      cmb_wr_data_b   <= (others=>'0');
      conv_ch_cnt     <= to_unsigned(c_CH_NUMBER, 4); --start with a temperature reading
      conv_wait_cnt_u <= c_START_DUR_US;
      state           <= s_STARTUP;

    elsif rising_edge(clk_ik) then
      if cen_ie = '1' then
        
        --disable
        if enable_i ='0' then
          cmb_wr_rd       <= '0';
          cmb_wr_data_b   <= (others=>'0');
          conv_ch_cnt     <= to_unsigned(c_CH_NUMBER, 4); --start with a temperature reading
          conv_wait_cnt_u <= c_START_DUR_US;
          state           <= s_STARTUP;
        else
      
          case state is
          
            -----------------
            -- wait at startup
            when s_STARTUP =>
                --go to IDLE
                if conv_wait_cnt_u = 0 then
                  conv_wait_cnt_u <= c_CONV_DUR_US;
                  state           <= s_IDLE;
                --wait at startup
                else
                  if timer_1us_i='1' then
                    conv_wait_cnt_u <= conv_wait_cnt_u - 1;
                  end if;
                end if;
          
            -----------------
            -- idle : wait for the start
            when s_IDLE =>
              cmb_wr_rd         <= '0';       
              cmb_wr_data_b     <= (others=>'0');
              if start_spi_cycle = '1' then     
                state           <= s_CONV;
              end if;
              
            -----------------
            -- conversion command: 
            when s_CONV =>
              if cmb_busy_i = '0' then
                cmb_wr_rd       <= '1';
                state           <= s_CONV_WAIT;
                --channel conversion 
                if conv_ch_cnt < c_CH_NUMBER then 
                                  -- CFG |  INCC | INx                                       | BW  |  REF  | SEQ  | RB_N | Dummy
                  cmb_wr_data_b   <= '1' & "111" & std_logic_vector(conv_ch_cnt(2 downto 0)) & '1' & "000" & "00" &  "1" & "00"; 
                --temperature conversion
                else --conv_ch_cnt = c_CH_NUMBER
                                  -- CFG |  INCC | INx  |  BW  |  REF  | SEQ  | RB_N | Dummy
                  cmb_wr_data_b   <= '1' & "011" & "000" & '1' & "000" & "00" &  "1" & "00";    --INCC: 011=temperature
                end if;
              end if;
              
            -----------------
            -- wait for the end of the conversion: 
            when s_CONV_WAIT =>
              if cmb_busy_i = '0' then
                --stop the cycle
                if conv_wait_cnt_u = 0 then
                  conv_wait_cnt_u <= c_CONV_DUR_US;
                  state           <= s_IDLE;
                  if conv_ch_cnt < c_CH_NUMBER then 
                    conv_ch_cnt   <= conv_ch_cnt+1;
                  else
                    conv_ch_cnt   <= (others=>'0');
                  end if;
                --wait for the end of the conversion
                else
                  cmb_wr_rd        <= '0';
                  if timer_1us_i='1' then
                    conv_wait_cnt_u <= conv_wait_cnt_u - 1;
                  end if;
                end if;
              end if;
          
            -----------------
            -- other state           
            when others =>
              state           <= s_IDLE;
          
          end case;
        end if;
      end if;
    end if;
  end process;

  -----------------
  -- SPI read valid to registers
  -----------------
  p_spi_command: process(clk_ik, rst_ir)
  begin
    if rst_ir = '1' then
      ADC3_LV_volt_valid_o  <= '0';
      ADC3_LV_volt_ob16     <= (others=>'0');
      ADC3_HV1_curr_valid_o <= '0';
      ADC3_HV1_curr_ob16    <= (others=>'0');
      ADC3_HV2_curr_valid_o <= '0';
      ADC3_HV2_curr_ob16    <= (others=>'0');
      ADC3_LV_curr_valid_o  <= '0';
      ADC3_LV_curr_ob16     <= (others=>'0');
      ADC3_4ch_temp_valid_o <= '0';
      ADC3_4ch_temp_ob16    <= (others=>'0');
    

    elsif rising_edge(clk_ik) then
      if cen_ie = '1' then
        ADC3_LV_volt_valid_o  <= '0';
        ADC3_HV1_curr_valid_o <= '0';
        ADC3_HV2_curr_valid_o <= '0';
        ADC3_LV_curr_valid_o  <= '0';
        ADC3_4ch_temp_valid_o <= '0';
        if cmb_rd_valid_i='1' then
          -- Channel 0: LV voltage
          if    conv_ch_cnt = 2 then --index shifted by 2, because the acquisition is 2 cycle after the conv 
            ADC3_LV_volt_ob16    <= cmb_rd_data_ib(15 downto 0);  
            ADC3_LV_volt_valid_o <= '1'; 
          -- Channel 1: HV1 current
          elsif conv_ch_cnt = 3 then
            ADC3_HV1_curr_ob16    <= cmb_rd_data_ib(15 downto 0); 
            ADC3_HV1_curr_valid_o <= '1';
          -- Channel 2: HV2 current
          elsif conv_ch_cnt = 4 then
            ADC3_HV2_curr_ob16    <= cmb_rd_data_ib(15 downto 0); 
            ADC3_HV2_curr_valid_o <= '1';
          -- Channel 3: LV current
          elsif conv_ch_cnt = 0 then
            ADC3_LV_curr_ob16    <= cmb_rd_data_ib(15 downto 0);  
            ADC3_LV_curr_valid_o <= '1';
          -- Temperature
          elsif conv_ch_cnt = 1 then
            ADC3_4ch_temp_ob16 <= cmb_rd_data_ib(15 downto 0); 
            ADC3_4ch_temp_valid_o <= '1';  
          else
            null;
          end if;
        end if;
      end if;
    end if;
  end process;

  -------------------
  -- CMB output
  -------------------
  cmb_frame_o    <= '0';           -- no frame
  cmb_addr_ob    <= "1";           -- 1 slave only (hot bit)
  cmb_enable_ob  <= (others=>'1'); -- all bits enable
  cmb_wr_o       <= cmb_wr_rd;     -- RD and WR in parallel (start the new conversion)
  cmb_wr_data_ob <= cmb_wr_data_b; 
  cmb_rd_o       <= cmb_wr_rd;     -- RD and WR in parallel (read the previsous acquisition)

end architecture;