--============================================================================================
--##################################   Module Information   ##################################
--============================================================================================
--
-- Company: CERN 
--
-- Language: VHDL 2008
--
-- Description:
--
--      See dac_ad5663armz Module
--
------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

------------------------------------------------------------------------------------------------------
-- Entity
------------------------------------------------------------------------------------------------------
entity dac_ad5663_spi_protocol is
  generic(
    -- Local slave bus
    g_CMB_SPI_ADDR_LEN    : positive := 1;  -- slave selection (hot bit)
    g_CMB_SPI_DATA_LEN    : positive := 24  -- SPI frame length in bits
  );
  port(
    --clock and reset
    rst_ir                : in  std_logic; 
    clk_ik                : in  std_logic; 
    cen_ie                : in  std_logic; 
    --enable
    enable_i              : in  std_logic; 
    --DAC CTRL              
    chA_wr_i              : in  std_logic; 
    chA_ib16              : in  std_logic_vector(15 downto 0); 
    chA_busy_o            : out std_logic; 
    chB_wr_i              : in  std_logic; 
    chB_ib16              : in  std_logic_vector(15 downto 0); 
    chB_busy_o            : out std_logic; 
    -- CMB SPI output
    cmb_frame_o           : out std_logic;
    cmb_addr_ob           : out std_logic_vector(g_CMB_SPI_ADDR_LEN-1 downto 0); --slave selection
    cmb_enable_ob         : out std_logic_vector(g_CMB_SPI_DATA_LEN-1 downto 0);
    cmb_wr_o              : out std_logic;
    cmb_wr_data_ob        : out std_logic_vector(g_CMB_SPI_DATA_LEN-1 downto 0); --shift reg content
    cmb_rd_o              : out std_logic;
    cmb_rd_data_ib        : in  std_logic_vector(g_CMB_SPI_DATA_LEN-1 downto 0);
    cmb_rd_valid_i        : in  std_logic;
    cmb_busy_i            : in  std_logic
  );
end entity ;


------------------------------------------------------------------------------------------------------
-- Architecture
------------------------------------------------------------------------------------------------------
architecture rtl of dac_ad5663_spi_protocol is
  
  --command and addresses
  constant c_CMD_WR_REG_UPDT : std_logic_vector(2 downto 0) := "011"; -- C2 C1 C0 = 011 : WR and update reg/ch n
  constant c_ADDR_CH_A       : std_logic_vector(2 downto 0) := "000"; -- A2 A1 A0 = 000 : DAC channel 1
  constant c_ADDR_CH_B       : std_logic_vector(2 downto 0) := "001"; -- A2 A1 A0 = 001 : DAC channel 2
  
  --CMB internal
  signal cmb_wr              : std_logic;
  signal cmb_wr_data_b       : std_logic_vector(g_CMB_SPI_DATA_LEN-1 downto 0);
  
  --Init
  signal init_chA_wr         : std_logic;
  signal init_chB_wr         : std_logic;
                             
begin 

  -------------------
  -- Initialisation
  -------------------
  p_init_dac: process(clk_ik, rst_ir)
  begin
    if rst_ir = '1' then
      init_chA_wr <= '1'; 
      init_chB_wr <= '0'; 

    elsif rising_edge(clk_ik) then
      if cen_ie = '1' then
        --disable
        if enable_i ='0' then
          init_chA_wr <= '1'; 
          init_chB_wr <= '0'; 
        else
          if cmb_busy_i = '0' then
            --Init chA after reset (with the input value)  
            if   init_chA_wr='1' then
              init_chA_wr <= '0';
              init_chB_wr <= '1';
            --Init chB after reset (with the input value)  
            elsif init_chB_wr='1' then
              init_chB_wr <= '0';
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;

  -------------------
  -- SPI command generation
  -------------------
  p_spi_command: process(clk_ik, rst_ir)
  begin
    if rst_ir = '1' then
      cmb_wr        <= '0';       
      cmb_wr_data_b <= (others=>'0');

    elsif rising_edge(clk_ik) then
      if cen_ie = '1' then
        --disable
        if enable_i ='0' then
          cmb_wr        <= '0';       
          cmb_wr_data_b <= (others=>'0');
        else
        
          if cmb_busy_i = '0' then
            --WR and update reg/ch A
            if chA_wr_i='1' OR init_chA_wr='1' then
              cmb_wr        <= '1';
              cmb_wr_data_b <= "00" & c_CMD_WR_REG_UPDT & c_ADDR_CH_A & chA_ib16;   
            --WR and update reg/ch B
            elsif chB_wr_i='1' OR init_chB_wr='1'then
              cmb_wr        <= '1';
              cmb_wr_data_b <= "00" & c_CMD_WR_REG_UPDT & c_ADDR_CH_B & chB_ib16;
            else
              cmb_wr        <= '0';       
              cmb_wr_data_b <= (others=>'0');
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;
  
  -------------------
  -- CMB output
  -------------------
  cmb_frame_o    <= '0';           -- no frame
  cmb_addr_ob    <= "1";           -- 1 slave only (hot bit)
  cmb_enable_ob  <= (others=>'1'); -- all 24b enable
  cmb_wr_o       <= cmb_wr;       
  cmb_wr_data_ob <= cmb_wr_data_b; 
  cmb_rd_o       <= '0';           --no read acces to the DAC
  
  -------------------
  -- Busy
  -------------------
  chA_busy_o <= cmb_busy_i or cmb_wr or not(enable_i);
  chB_busy_o <= cmb_busy_i or cmb_wr or not(enable_i);
 

end architecture;